// From http://www.redblobgames.com/grids/hexagons/
// Copyright 2015 Red Blob Games <redblobgames@gmail.com>
// License: Apache v2.0 <http://www.apache.org/licenses/LICENSE-2.0.html>
//
// Helper functions used by Output* classes

using Ast;
using StringTools;
import Codegen;

class Output {
    public static function indent(n:Int) {
        return "".lpad(' ', 4 * n);
    }

    
    public static function floatLiteral(n:Float) {
        // I want to make sure it doesn't look like an int literal
        var s = Std.string(n);
        if (s.indexOf(".") < 0 && s.indexOf("e") < 0) { s = s + ".0"; }
        return s;
    }

    
    public static function stringLiteral(s:String) {
        // I only use string literals for test names, so I'm not going to deal with escaping
        var allowed = new EReg("^[^\'\"]*$", "");
        if (allowed.match(s)) {
            return '"${s}"';
        } else {
            return "#invalid-string";
        }
    }

    
    // Convert "abc" to "Abc"
    public static function capitalize(s:String) {
        return if (s.length > 0) s.substr(0, 1).toUpperCase() + s.substr(1) else s;
    }

    
    // Convert "abc_qrs_xyz" to "AbcQrsXyz"
    public static function nameToPascalCase(name:String) {
        var words = name.split("_");
        var capWords = [for (w in words) capitalize(w)];
        return capWords.join("");
    }

    
    // Convert "abc_qrs_xyz" to "abcQrsXyz"
    public static function nameToCamelCase(name:String) {
        var words = name.split("_");
        var capWords = [for (w in words.slice(1)) capitalize(w)];
        return [words[0]].concat(capWords).join("");
    }


    // Merge 'base' and 'adopt' classes.
    // 1. Keep fields only from base class.
    // 2. Keep methods, static vars from both classes.
    // 3. Keep fail condition only for adopted class.
    // NOTE: keeping the adopted invariant may not be the
    // right thing in general; it happens to be useful for
    // Hex + FractionalHex which is the only time we call this
    public static function adoptFunctions(base:Array<DExpr>, adopt:Array<DExpr>) {
        var keep = [];
        for (d in base) {
            switch (d) {
                case DPrecondition(_, _): /* skip */ null;
                case _: keep.push(d);
            }
        }
        for (d in adopt) {
            switch (d) {
                case DField(_, _): /* skip */ null;
                case _: keep.push(d);
            }
        }
        return keep;
    }

    // Gather all the fields if they are to be placed somewhere
    // other than the main class definition
    public static function collectFields(decls:Array<DExpr>):Array<{name:String, type:TExpr}> {
        var results = [];
        for (d in decls) {
            switch (d) {
                case DField(name, type): results.push({name: name, type: type});
                case _:
            }
        }
        return results;
    }


    // Gather the failure conditions so they can be placed in the constructor
    public static function collectPreconditions(decls:Array<DExpr>):Null<{test: AExpr, message: AExpr}> {
        for (d in decls) {
            switch (d) {
                case DPrecondition(test, message): return {test: test, message: message};
                case _:
            }
        }
        return null;
    }
        
    
    // Default type names; subclasses can override these for language-specific names
    function Type(t:TExpr) {
        return switch (t) {
            case TVoid: 'Void';
            case TInt: 'Int';
            case TFloat: 'Float';
            case TString: 'String';
            case THex: 'Hex';
            case TFractionalHex: 'FractionalHex';
            case TOffsetCoord: 'OffsetCoord';
            case TDoubledCoord: 'DoubledCoord';
            case TPoint: 'Point';
            case TLayout: 'Layout';
            case TOrientation: 'Orientation';
            case TArray(p): 'Array<${Type(p)}>';
        };
    }


    // Default expr output; most of the Algol-ish outputs have some similarity
    function Expr(e:AExpr) {
        return switch (e) {
            case AInt(num): '$num';
            case AFloat(num): floatLiteral(num);
            case AString(str): stringLiteral(str);
            case AArrayIndex(a, i): '${Expr(a)}[${Expr(i)}]';
            case AFieldAccess(a, i): '$a.$i';
            case AParenthesis(expr): '(${Expr(expr)})';
            case _: ':TODO:';
        }
    }

    
    public static function outputAll(output, unifiedNumberType) {
        Sys.println(output('Point', Codegen.pointClass()));
        if (unifiedNumberType) {
            Sys.println(output('Hex', adoptFunctions(Codegen.hexClass(),
                                               Codegen.fractionalHexClass())));
        } else {
            Sys.println(output('Hex', Codegen.hexClass()));
            Sys.println(output('FractionalHex', Codegen.fractionalHexClass()));
        }
        Sys.println(output('OffsetCoord', Codegen.offsetClass()));
        Sys.println(output('DoubledCoord', Codegen.doubledClass()));
        Sys.println(output('Orientation', Codegen.orientationClass()));
        Sys.println(output('Layout', Codegen.layoutClass()));
    }

    public static function outputTests(output) {
        Sys.println(output('Tests', Codegen.testsClass()));
    }
}
