// From http://www.redblobgames.com/
// Copyright 2015 Red Blob Games <redblobgames@gmail.com>
// License: Apache v2.0 <http://www.apache.org/licenses/LICENSE-2.0.html>

//
// Haxe output

using Ast;
using StringTools;
import Output.*;

class OutputHaxe extends Output {
    
    function new() {}

    // Rename functions Cube.cube_* to Cube.*, Hex.hex_* to Hex.*, etc.
    function rename(context:String, name:String, isMethod:Bool) {
        var prefix = context.toLowerCase() + "_";
        if (name.startsWith(prefix)) { name = name.substr(prefix.length); }
        if (isMethod) { name = Output.nameToCamelCase(name); }
        return name;
    }

    // Build up a mapping of all the renames, so that when emitting
    // function calls we know what those functions are now called and
    // how to now call them.
    var nameMapping:Map<String, {context:String, name:String}> = new Map();
    function collectNames(className:String, decls:Array<DExpr>) {
        for (decl in decls) {
            switch(decl) {
                case DStaticVar(name, _, _):
                    nameMapping[name] = {context: className, name: rename(className, name, false)};
                case DMethod(name, _, _, _):
                    nameMapping[name] = {context: className, name: rename(className, name, true)};
                case _:
            }
        }
        
        return "";
    }

    
    override function Expr(e:AExpr) {
        return switch (e) {
            case AName(name):
                switch (name) {
                    case "PI": "Math.PI";
                    case _: if (nameMapping.exists(name)) '${nameMapping[name].context}.${nameMapping[name].name}' else name;
                }
            case ANew(TArray(t), arguments): '[${[for (a in arguments) Expr(a)].join(", ")}]';
            case ANew(type, arguments): 'new ${Type(type)}(${[for (a in arguments) Expr(a)].join(", ")})';
            case AFunctionCall("#negate", [expr]): '-${Expr(expr)}';
            case AFunctionCall("#not", [expr]): '!${Expr(expr)}';
            case AFunctionCall("Int", [AFunctionCall("round", [expr])]): Expr(AFunctionCall("round", [expr])); // In Haxe, round already returns int
            case AFunctionCall("$error", [expr]): 'throw ${Expr(expr)}';
            case AFunctionCall("$array_length", [AName(a)]): Expr(AFieldAccess(a, "length"));
            case AFunctionCall("#range", [a, b]): '${Expr(a)}...${Expr(b)}';
            case AFunctionCall("Float", [expr]): Expr(expr);
            case AFunctionCall(name, arguments):
                if ((name == ">" || name == "==" || name == "!=" || name == "+" || name == "-" || name == "*" || name == "/" || name == "and" || name == "&") && arguments.length == 2) {
                    if (name == "and") { name = "&&"; }
                    '${Expr(arguments[0])} $name ${Expr(arguments[1])}';
                } else {
                    if (name == "Int") { name = "Std.int"; }
                    if (name == "sin") { name = "Math.sin"; }
                    if (name == "cos") { name = "Math.cos"; }
                    if (name == "round") { name = "Math.round"; }
                    if (name == "abs") { name = "Math.abs"; }
                    if (name == "sqrt") { name = "Math.sqrt"; }
                    if (name == "max") { name = "Math.max"; }
                    if (nameMapping.exists(name)) { name = '${nameMapping[name].context}.${nameMapping[name].name}'; }
                    '$name(${[for (a in arguments) Expr(a)].join(", ")})';
                }
            case AMethodCall(name, method, arguments):
                '$name.$method(${[for (a in arguments) Expr(a)].join(", ")})';
            case _: super.Expr(e);
        };
    }

                           
    function Stmt(level:Int, e:AExpr) {
        return switch (e) {
            case ABlock(exprs): '${indent(level-1)}{\n' + [for (e in exprs) Stmt(level, e)].join("") + '${indent(level-1)}}\n';
            case AFor(name, range, body): '${indent(level)}for ($name in ${Expr(range)})\n${Stmt(1+level, body)}';
            case AIf(test, ifTrue, ABlock([])): '${indent(level)}if (${Expr(test)})\n${Stmt(1+level, ifTrue)}';
            case AIf(test, ifTrue, ifFalse): '${indent(level)}if (${Expr(test)})\n${Stmt(1+level, ifTrue)}${indent(level)}else\n${Stmt(1+level, ifFalse)}';
            case AFunctionCall("set!", [AName(name), expr]): '${indent(level)}$name = ${Expr(expr)};\n';
            case AFunctionCall("set!", [AArrayIndex(AName(name), index), expr]): '${indent(level)}$name[${Expr(index)}] = ${Expr(expr)};\n';
            case AVar(name, type, init): '${indent(level)}var $name:${Type(type)} = ${Expr(init)};\n';
            case AReturn(expr): '${indent(level)}return ${Expr(expr)};\n';
            case _: '${indent(level)}${Expr(e)};\n';
        };
    }

                                      
    function Decl(context:String, decl:DExpr) {
        return switch(decl) {
            case DField(name, type):
                '${indent(1)}public var $name:${Type(type)};\n';
            case DStaticVar(name, type, init):
                '${indent(1)}static public var ${rename(context, name, false)}:${Type(type)} = ${Expr(init)};\n';
            case DMethod(name, type, params, body):
                '\n${indent(1)}static public function ${rename(context, name, true)}(${[for (p in params) p.name + ":" + Type(p.type)].join(", ")}):${Type(type)}\n${Stmt(2, body)}\n';
            case _: "";
        };
    }

                              
    function output(className:String, decls:Array<DExpr>) {
        var fields = collectFields(decls);
        var preconditions = collectPreconditions(decls);
        return ['class ${className}\n{\n',
                '${indent(1)}public function new(',
                [for (f in fields) '${f.name}:${Type(f.type)}'].join(", "),
                ")\n",
                '${indent(1)}{\n',
                [for (f in fields) '${indent(2)}this.${f.name} = ${f.name};\n'].join(""),
                (preconditions != null)? '${indent(2)}if (${Expr(preconditions.test)}) throw ${Expr(preconditions.message)};\n' : '',
                '${indent(1)}}\n'
                ].join("")
            + [for (decl in decls) Decl(className, decl)].join("")
            + "}\n";
    }
    

    public static function main() {
        var output = new OutputHaxe();
        
        Sys.println('// Generated code -- CC0 -- No Rights Reserved -- http://www.redblobgames.com/grids/hexagons/
');

        Output.outputAll(output.collectNames, false);
        Output.outputAll(output.output, false);

        Sys.println('

// Tests

');

        var testsClass = Codegen.testsClass();
        testsClass.push(DMethod("main", TVoid, [], ABlock([AFunctionCall("test_all", [])])));
        testsClass.push(DMethod("complain", TVoid, [{name: "name", type: TString}], ABlock([AFunctionCall("trace", [AString("FAIL "), AName("name")])])));
        output.collectNames('Tests', testsClass);
        Sys.println(output.output('Tests', testsClass));
    }
}
